import json
if __name__ == "__main__":
    from generateGeneralLanguajeModel import generateGeneralLanguajeModel
else:
    from speako.source.generateGeneralLanguajeModel import generateGeneralLanguajeModel


#Script que crea los modelos de lenguaje a partir de los archivos (.txt o .json) proporcionados

#Se especifican los archivos con los que crear los modelos de lenguaje en dos listas: fileFormat (formato de los archivos) y file (archivo)
#Como ejemplo, se propone un archivo .txt y otro .json
fileFormat = []
file = []
fileFormat.append('txt')
file.append('frasesTest.txt')
fileFormat.append('json')
file.append('test.json')

#Luego, se inicializan las listas y diccionarios a generar. Son los siguientes:
listWords = ['.', '?','!'] #Lista de palabras incluídas en el diccionario, inicializada con los signos de puntuación de fin de frase
listWordsInf = ['.', '?','!'] #Lista de palabras incluídas en el diccionario de infinitivos, inicializada con los signos de puntuación de fin de frase
dictNgram = {} #Diccionario de N-gramas
dictNgramInf = {} #Diccionario de N-gramas con verbos en infinitivo
dictBackoff = {} #Diccionario de parámetro backoff
dictBackoffInf = {} #Diccionario de parámetro backoff con verbos en infinitivo

for i in range(len(file)): #Para todos los archivos especificados, se generan los modelos de lenguaje añadiendo los datos recursivamente
	[listWords, listWordsInf, dictNgram, dictNgramInf, dictBackoff, dictBackoffInf] = generateGeneralLanguajeModel(fileFormat[i], file[i], listWords, listWordsInf, dictNgram, dictNgramInf, dictBackoff, dictBackoffInf)


#Se guardan todos los datos (listas y diccionarios), cada uno en un archivo .txt
f = open("listWords.txt", "w")
for word in listWords:
    f.write(word+'\n')
f.close()

f = open("listWordsInf.txt", "w")
for word in listWordsInf:
    f.write(word+'\n')
f.close()

f = open("dictNgram.txt", "w")
json.dump(dictNgram, f)
f.close()

f = open("dictNgramInf.txt", "w")
json.dump(dictNgramInf, f)
f.close()

f = open("dictBackoff.txt", "w")
json.dump(dictBackoff, f)
f.close()

f = open("dictBackoffInf.txt", "w")
json.dump(dictBackoffInf, f)
f.close()


if __name__ == '__main__':
	from source.calcularPrediccionCache import calcularPrediccionCache
	from source.calcularPrediccionMultPal import calcularPrediccion
	from source.generateCacheLanguajeModel import generateCacheLanguajeModel
else:
	from speako.source.calcularPrediccionCache import calcularPrediccionCache
	from speako.source.calcularPrediccionMultPal import calcularPrediccion
	from speako.source.generateCacheLanguajeModel import generateCacheLanguajeModel
import json
import os

try:
    import importlib.resources as pkg_resources
except ImportError:
    import importlib_resources as pkg_resources

class predictor:

	# --- Private Methods ---
	_inCaseNoDefaultPathsIsFound = "quiero, hola, yo, estás, comer,\
									la, tengo, con, a, jugar"
	def __init__(self, model = "base", model_dir = "./", language = "es"):
		"""
		Description:
		initialization method

		Inputs:
		- model (str, optional): The name of the model that's going to be used for the prediction.
		There should be a folder with the same name as this on the model's directory that
		contains the corresponding files for this model. Default is "base".
		- model_dir (str, optional): The absolute path to the directory where the model and language files
		are stored. If they are located where the model is being executed from this parameter can be
		left as default.
		- language (str, optional): This is the language that will be used for the prediction. On
		the model's directory there should be a folder name "verb-conj" with the verb conjugations for each
		language named "lang-verbs.txt" where "lang" is a short string representing each language: "es" for 
		spanish, "en" for english and so.

		Outputs:
		predictor (Instance of predictor Class)
		"""
		self._model = model
		if model_dir == "":
			self._model_dir = os.getcwd()
		else:
			self._model_dir = model_dir
		self._language = language

		self._load_model()

	def _load_model(self):
		(self._listWords, self._listWordsInf) = self._load_listWords()
		(self._dictNgram, self._dictNgramInf
		, self._dictBackoff, self._dictBackoffInf)  = self._load_dicts()
		(self._listWordsCache, self._listWordsInfCache) = self._load_listWords(is_cache = True)
		(self._dictNgramCache, self._dictNgramInfCache
		, self._dictBackoffCache, self._dictBackoffInfCache)  = self._load_dicts(is_cache = True)
		self._verbsArray = self._load_language()
		self._defaultPaths = self._load_def_paths()

	def _load_listWords(self, is_cache = False):
		if is_cache:
			scan_dir = os.path.join(self._model_dir, "cache")
			print("Loading files from:", scan_dir)
			try:
				f = open(os.path.join(scan_dir, "listWordsCache.txt"), "r")
				content = f.read()
				lw = content.split("\n")
				f.close()

				f = open(os.path.join(scan_dir, "listWordsInfCache.txt"), "r")
				content = f.read()
				lwi = content.split("\n")
				f.close()
			except FileNotFoundError:
				lw, lwi = [".", "?", "!"], [".", "?", "!"] 
		else:
			scan_dir = os.path.join(self._model_dir, self._model)
			print("Loading files from:", scan_dir)
			try:
				f = open(os.path.join(scan_dir, "listWords.txt"), "r")
				content = f.read()
				lw = content.split("\n")
				f.close()

				f = open(os.path.join(scan_dir, "listWordsInf.txt"), "r")
				content = f.read()
				lwi = content.split("\n")
				f.close()
			except FileNotFoundError:
				print("The model you're trying to use does not exist, please check the model name or models directory.")
				lw, lwi = None, None
		return lw, lwi

	def _load_dicts(self, is_cache = False):
		if is_cache:
			scan_dir = os.path.join(self._model_dir, "cache")
			print("Loading files from:", scan_dir)
			try:
				f = open(os.path.join(scan_dir, "dictNgramCache.txt"), "r")
				dn = json.load(f)
				f.close()
				#nTot = list(dictNgram.values())[0][3]#Tamaño del diccionario de N-gramas principal

				f = open(os.path.join(scan_dir, "dictNgramInfCache.txt"), "r")
				dni = json.load(f)
				f.close()

				f = open(os.path.join(scan_dir, "dictBackoffCache.txt"), "r")
				db = json.load(f)
				f.close()

				f = open(os.path.join(scan_dir, "dictBackoffInfCache.txt"), "r")
				dbi = json.load(f)
				f.close()
			except FileNotFoundError:
				dn, dni, db, dbi = {}, {}, {}, {}
		else:
			scan_dir = os.path.join(self._model_dir, self._model)
			print("Loading files from:", scan_dir)
			try:
				f = open(os.path.join(scan_dir, "dictNgram.txt"), "r")
				dn = json.load(f)
				f.close()
				#nTot = list(dictNgram.values())[0][3]#Tamaño del diccionario de N-gramas principal

				f = open(os.path.join(scan_dir, "dictNgramInf.txt"), "r")
				dni = json.load(f)
				f.close()

				f = open(os.path.join(scan_dir, "dictBackoff.txt"), "r")
				db = json.load(f)
				f.close()

				f = open(os.path.join(scan_dir, "dictBackoffInf.txt"), "r")
				dbi = json.load(f)
				f.close()
			except FileNotFoundError:
				print("The model you're trying to use does not exist, please check the model name or models directory.")
				dn, dni, db, dbi = None, None, None, None
		return dn, dni, db, dbi

	def _load_language(self):

		try:
			if __name__ == '__main__':
				scan_path = os.path.join("lang", f"{self._language}-verbs.txt")
				verb_file = open(scan_path, encoding="utf-8") #Se abre el archivo que contiene todas las formas verbales de los verbos en español
			else:
				from . import lang
				verb_file = pkg_resources.open_text(lang, f"{self._language}-verbs.txt")
			va = verb_file.read().split()
		except FileNotFoundError:
			print("""The language input does not exist in our database, check if the input was
				right or raise an issue on our github page in case it's not loaded yet.""")
			va = None
		return va

	def _load_def_paths(self):
		scan_path = os.path.join(self._model_dir, "listDefaultPaths.txt")
		print("Loading files from:", scan_path)
		try:
			verb_file = open(scan_path, "r", encoding="utf-8") 
			dp = verb_file.read().split()
		except FileNotFoundError:
			print("Could not find a default path file, so we'll set the default paths to:")
			print(self._inCaseNoDefaultPathsIsFound)
			dp = self._inCaseNoDefaultPathsIsFound.split(", ")
		return dp

	def _is_cache_initial(self):
		return any((self._listWordsCache == [".", "?", "!"],
					self._listWordsInfCache == [".", "?", "!"],
					self._dictNgramCache == {},
					self._dictNgramInfCache == {},
					self._dictBackoffCache == {},
					self._dictBackoffInfCache == {},
				  ))

	def change_model_params(self, **kwargs):
		"""
		Description: 
		method to change the parameters used for the prediction.

		Inputs:
		- model (str, optional): The name of the model that's going to be used for the prediction.
		There should be a folder with the same name as this on the model's directory that
		contains the corresponding files for this model. If not passed, the model name stays the same.
		- model_dir (str, optional): The absolute path to the directory where the model and language files
		are stored. If they are located where the model is being executed from this parameter can be
		passed a "./". If not passed, the model dir stays the same.
		- language (str, optional): This is the language that will be used for the prediction. On
		the model's directory there should be a folder name "verb-conj" with the verb conjugations for each
		language named "lang-verbs.txt" where "lang" is a short string representing each language: "es" for 
		spanish, "en" for english and so. If not passed, the model language stays the same.

		Outputs:
		None
		"""
		for k in kwargs.keys():
			if k == "model":
				self._model = kwargs[k]
			elif k == "model_dir":
				self._model_dir = kwargs[k]
			elif k == "language":
				self._language = kwargs[k]
			else:
				print(f'{k} is not a valid parameter.')
		self._load_model()

	def get_model_description(self):
		"""
		Description:
		logs the model parameters including model name, directory and language, which
		can be changed with the change_model_params method.

		Inputs:
		None

		Outputs:
		None
		"""
		print(f'The currently load model is:\n \
				\t-> model name: {self._model}\n \
				\t-> model directory: {self._model_dir}\n \
				\t-> language: {self._language}')

	def predict(self, sentence = "", maxNgram = 5, maxNumPaths = 4, maxNumWords = 1, cacheAlfa = 0.6):
		"""
		Description:
		method in charge of carrying out the prediction on a certain phrase and returning a dictionary with
		the predicted options as output. Other than "sentence" which takes the partial phrase to make
		predictions on, other parameters control how the prediction will be done.

		Inputs:
		- sentence (str, optional): partial phrase to make predictions on. Default is empty phrase, where
		the model will provide the default starting paths.
		- maxNgram (int, optional): max amount of words from last to first the model will use for the
		predictions. For example, for the phrase "I want a candy" with maxNgram=3 the model will use "candy", "a candy"
		and "want a candy" as bases for the prediction. The higher this value is the more accurate the 
		prediction will be, but the longer it'll take. Default is 5.
		- maxNumPaths (int, optional): max possible options the model will offer predictions for. The higher
		this number is, the more options for continuing the phrase the model might predict, but the longer it'll take.
		Default is 4.
		- maxNumWords (int, optional): max amount of words the algorithm will try to go for on a each prediction.
		For example, with a maxNumWords=3, for the phrase "Hi" the model might predict "what's up" (2 words) or
		"my name is" (3 words), but with maxNumWords=1 it'll predict "my" or "what's".
		- cacheAlfa (float, optional): ratio, between 0 and 1, of the prediction that will come from the cache database.

		Outputs:
		-jsonPathsConjugated (dict): dictionary with the number of prediction as keys and the predictions as values. 
		"""
		if self._is_cache_initial():
			json_pathsConjugated = calcularPrediccion(sentence, maxNgram, maxNumPaths, maxNumWords, 
													self._listWords, self._listWordsInf, 
													self._dictNgram, self._dictNgramInf, 
													self._dictBackoff, self._dictBackoffInf, 
													self._verbsArray, self._defaultPaths
													)
		else:
			json_pathsConjugated = calcularPrediccionCache(sentence, maxNgram, maxNumPaths, maxNumWords, 
													self._listWords, self._listWordsInf, 
													self._dictNgram, self._dictNgramInf, 
													self._dictBackoff, self._dictBackoffInf,
													self._listWordsInfCache, 
													self._dictNgramCache, self._dictNgramInfCache, 
													self._dictBackoffInfCache, 
													self._verbsArray, self._defaultPaths,
													cacheAlfa
													)

		return json_pathsConjugated

	def update_cache(self, sentence):
		"""
		Description:
		method dedicated to updating the data in the cache folder on the model's directory
		(or initializing it for the first time). This cache data works as a separate model that
		complements the prediction made by the main model to adapt it to each particular user.
		When a sentence is passed as a parameter, the cache model will be updated on said phrase,
		making it more likely to appear as a prediction on following predictions, even when
		the main model is not updated on that phrase.

		Inputs:
		- sentence (str): sentence the cache model will be updated on.

		Outputs:
		None
		"""
		cache_dir = os.path.join(self._model_dir, "cache")
		sen_json_path = os.path.join(cache_dir, "newSentenceCache.json")
		f = open(sen_json_path, "w")
		sen_dict = {"frase": sentence}
		json_data = json.dumps(sen_dict) #Se convierte el diccionario en un string JSON
		json_paths = json.loads(json_data) # Se crea el objeto JSON
		f.write("[")
		json.dump(json_paths,f)
		f.write("]")
		f.close()
		(self._listWordsCache, self._listWordsInfCache, 
		self._dictNgramCache, self._dictNgramInfCache, 
		self._dictBackoffCache, self._dictBackoffInfCache) = generateCacheLanguajeModel("json", sen_json_path, 
																				self._listWordsCache, self._listWordsInfCache, 
																				self._dictNgramCache, self._dictNgramInfCache, 
																				self._dictBackoffCache, self._dictBackoffInfCache, 
																				self._verbsArray)

		f = open(os.path.join(cache_dir, "listWordsCache.txt"), "w")
		for word in self._listWordsCache:
			if word == self._listWordsCache[len(self._listWordsCache)-1]:
				f.write(word)
			else:
				f.write(word+'\n')
		f.close()

		f = open(os.path.join(cache_dir, "listWordsInfCache.txt"), "w")
		for word in self._listWordsInfCache:
			if word == self._listWordsInfCache[len(self._listWordsInfCache)-1]:
				f.write(word)
			else:
				f.write(word+'\n')
		f.close()

		f = open(os.path.join(cache_dir, "dictNgramCache.txt"), "w")
		json.dump(self._dictNgramCache, f)
		f.close()

		f = open(os.path.join(cache_dir, "dictNgramInfCache.txt"), "w")
		json.dump(self._dictNgramInfCache, f)
		f.close()

		f = open(os.path.join(cache_dir, "dictBackoffCache.txt"), "w")
		json.dump(self._dictBackoffCache, f)
		f.close()

		f = open(os.path.join(cache_dir, "dictBackoffInfCache.txt"), "w")
		json.dump(self._dictBackoffInfCache, f)
		f.close()
	
	#Missing: adding updating cache to interactive, since now it is crashing
	def interactive_predict(self, sentence = "", maxNgram = 5, maxNumPaths = 4, maxNumWords = 1, 
							cacheAlfa = 0.6, cache_update = "none"):
		"""
		Description:
		this is method implements a basic example of an application made with
		the help of this class, where the sentence and params are passed to the predictor,
		the results of the prediction are shown and the user input is solicitated to choose
		one, which is used then to build up on the sentence and run the prediction again, repeating
		the process untill the user stops it with and input of 0.

		Inputs:
		cache_update (str, optional): this can be "all", to update cache after every prediction;
		"final" to update it only when the user finishes the sentence building, or "none" for the cache not to
		be updated by this method. Default is "none".
		See predictor.predict method for the description of the parameters.

		Outputs:
		final (str): when the user inputs a 0 the process is terminated and the final sentence
		is returned as a string.
		"""

		if all([cache_update != s for s in ["all", "final", "none"]]):
			raise ValueError(f'{cache_update} is not a valid option for cache_update. See doc to check available options.')

		choice = -1
		paths = self.predict(sentence, maxNgram , maxNumPaths, maxNumWords, cacheAlfa)
		print(f'Prediction for {sentence}:')
		print(paths)
		while choice != 0:
			try:
				choice = int(input("Enter your choice, or 0 to end prediction: "))
				if 0 < choice < len(paths) + 1:
					sentence += (" " + paths[str(choice)]) 
					paths = self.predict(sentence, maxNgram , maxNumPaths, maxNumWords, cacheAlfa)
					print(f'Prediction for {sentence}:')
					print(paths)
					if cache_update == "all":
						self.update_cache(sentence)
				elif choice == 0:
					print(f'Sentence ended as {sentence}')
					if cache_update == "final":
						self.update_cache(sentence)
					return sentence
				else:
					print(f'Invalid input, should be a value greater or equal than zero and lesser or equal than {len(paths)}.')
			except ValueError:
				print(f'Invalid input, should be a value greater or equal than zero and lesser or equal than {len(paths)}.')

if __name__ == "__main__":

	pred = predictor(model = "med", model_dir="C:/Users/Juanma/OTTAA - Lixi/PictogramsPredictionsLibrary/")
	final = pred.interactive_predict("Quiero", cache_update = "all")
	print(final)


if __name__ == '__main__':
    from source.generateGeneralLanguajeModel import generateGeneralLanguajeModel
else:
    from speako.source.generateGeneralLanguajeModel import generateGeneralLanguajeModel
import json
import os
import statistics as stats
import random

try:
    import importlib.resources as pkg_resources
except ImportError:
    import importlib_resources as pkg_resources

class trainer:

    def __init__(self, data_path, data_format = "txt", model = "base", model_dir = "./", language = "es"):
        """
		Description:
		initialization method

		Inputs:
        - data_path (str): The absolute path for the file with the phrases that will be used as data.
        The class also provides a method to train with multiple files at once, in case that's desired then
        pass any path to this parameter and provide the correct path for the multiple files directory
        when the train_multiple method is called.
        - data_format (str, optional): one of "json", "txt" or "any". Ensurance that the file is 
        correctly loaded, the data format of the file will be detected and compared to this parameter. 
        It can be one of two: "json" or "txt". Pass "any" if you don't want any checking of data format. 
        Specially important when using the multiple_train method.
		- model (str, optional): The name of the model that's going to be used for the prediction.
		There should be a folder with the same name as this on the model's directory that
		contains the corresponding files for this model. Default is "base".
		- model_dir (str, optional): The absolute path to the directory where the model and language files
		are stored. If they are located where the model is being executed from this parameter can be
		left as default.
		- language (str, optional): This is the language that will be used for the prediction. On
		the model's directory there should be a folder name "verb-conj" with the verb conjugations for each
		language named "lang-verbs.txt" where "lang" is a short string representing each language: "es" for 
		spanish, "en" for english and so.

		Outputs:
		predictor (Instance of predictor Class)
		"""
        self._data_path = data_path
        self._data_format = data_format
        self._model = model
        self._model_dir = model_dir
        self._language = language

        #Initialization of model files data
        self._listWords = ['.', '?','!'] 
        self._listWordsInf = ['.', '?','!'] 
        self._dictNgram = {} 
        self._dictNgramInf = {} 
        self._dictBackoff = {} 
        self._dictBackoffInf = {}

        self._verbsArray = self._load_language()

    def _model_exists(self):
        """
        Description:
        method to check if the defined model already exists
        """
        return any([self._model == m for m in os.listdir(self._model_dir)])
        
    def _model_loaded(self):
        """
        Description:
        checks if the model files have already been loaded (true) of if they are on initialized mode (false)
        """
        return not all((self._listWords == [".", "?", "!"],
					    self._listWordsInf == [".", "?", "!"],
					    self._dictNgram == {},
					    self._dictNgramInf == {},
					    self._dictBackoff == {},
					    self._dictBackoffInf == {},
				        ))

    def _model_status(self):
        """
        Description
        method for checking the current status of the model, and asking the user if they want to continue
        or stop the training process and change some stuff before training again
        """
        if self._model_loaded():
            while True:
                print("There's already a model loaded into the instance, will continue training data on it.")
                print("If you were trying to train a completely new model stop this process and move this model folder from the model directory or change the model parameter.")
                ans = str(input("Keep training? [[y]/n]: "))
                if ans == "y" or ans == "":
                    print("Starting training on already loaded model.")
                    break
                elif ans == "n":
                    raise KeyboardInterrupt("Training process stopped.")
                else:
                    print("Invalid option.")
        else:
            if self._model_exists():
                while True:
                    print("Another model with the same name was found on the models directory.")
                    print("If you continue the training this model will be loaded and trained on.")
                    print("If you were trying to train a completely new model stop this process and move this model folder from the model directory or change the model parameter.")
                    ans = str(input("Keep training? [[y]/n]: "))
                    if ans == "y" or ans == "":
                        self._load_model()
                        print("Model loaded, starting training on it.")
                        break
                    elif ans == "n":
                        raise KeyboardInterrupt("Training process stopped.")
                    else:
                        print("Invalid option.")
            else:
                print("Starting training with a brand new model.")


    def _load_model(self):
        """
        Description:
        method for loading the model files
        """
        load_dir = os.path.join(self._model_dir, self._model)
        try:
            f = open(os.path.join(load_dir, "listWords.txt"), "r")
            content = f.read()
            self._listWords = content.split("\n")
            f.close()
            
            f = open(os.path.join(load_dir, "listWordsInf.txt"), "r")
            content = f.read()
            self._listWordsInf = content.split("\n")
            f.close()
            
            f = open(os.path.join(load_dir, "dictNgram.txt"), "r")
            self._dictNgram = json.load(f)
            f.close()
            
            f = open(os.path.join(load_dir, "dictNgramInf.txt"), "r")
            self._dictNgramInf = json.load(f)
            f.close()
            
            f = open(os.path.join(load_dir, "dictBackoff.txt"), "r")
            self._dictBackoff = json.load(f)
            f.close()
            
            f = open(os.path.join(load_dir, "dictBackoffInf.txt"), "r")
            self._dictBackoffInf = json.load(f)
            f.close()
        
        except FileNotFoundError as e:
            print("The model loading process failed because: ", e)
            print("Will set the model data to a new instance.")
            self._listWords = ['.', '?','!'] 
            self._listWordsInf = ['.', '?','!'] 
            self._dictNgram = {} 
            self._dictNgramInf = {} 
            self._dictBackoff = {} 
            self._dictBackoffInf = {}

    def _train_model(self, train_file, train_file_format):
        """
        Description:
        method for carrying out the training process, it's done sepparately to reuse it for both the train and train multiple.
        """
        [self._listWords, self._listWordsInf, 
        self._dictNgram, self._dictNgramInf, 
        self._dictBackoff, self._dictBackoffInf] = generateGeneralLanguajeModel(train_file_format, train_file, 
                                                                                self._listWords, self._listWordsInf, 
                                                                                self._dictNgram, self._dictNgramInf, 
                                                                                self._dictBackoff, self._dictBackoffInf, 
                                                                                self._verbsArray)
    
    def _save_model(self):
        """
        Description:
        method for saving the model files
        """
        save_dir = os.path.join(self._model_dir, self._model)

        f = open(os.path.join(save_dir, "listWords.txt"), "w")
        for word in self._listWords[:-1]:
            f.write(word+'\n')
        f.close()
        f = open(os.path.join(save_dir, "listWords.txt"), "a")
        f.write(self._listWords[-1])
        f.close()
        
        f = open(os.path.join(save_dir, "listWordsInf.txt"), "w")
        for word in self._listWordsInf[:-1]:
            f.write(word+'\n')
        f.close()
        f = open(os.path.join(save_dir, "listWordsInf.txt"), "a")
        f.write(self._listWordsInf[-1])
        f.close()
        
        f = open(os.path.join(save_dir, "dictNgram.txt"), "w")
        json.dump(self._dictNgram, f)
        f.close()
        
        f = open(os.path.join(save_dir, "dictNgramInf.txt"), "w")
        json.dump(self._dictNgramInf, f)
        f.close()
        
        f = open(os.path.join(save_dir, "dictBackoff.txt"), "w")
        json.dump(self._dictBackoff, f)
        f.close()
        
        f = open(os.path.join(save_dir, "dictBackoffInf.txt"), "w")
        json.dump(self._dictBackoffInf, f)
        f.close()

    def _load_language(self):
        """
        Description:
        method to load the language from
        """
        try:
            if __name__ == '__main__':
                scan_path = os.path.join("lang", f"{self._language}-verbs.txt")
                verb_file = open(scan_path, encoding="utf-8") #Se abre el archivo que contiene todas las formas verbales de los verbos en español
            else:
                from . import lang
                verb_file = pkg_resources.open_text(lang, f"{self._language}-verbs.txt")
            va = verb_file.read().split()
        except FileNotFoundError:
            print("""The language input does not exist in our database, check if the input was
            right or raise an issue on our github page in case it's not loaded yet.""")
            va = None
        return va

    def change_model_params(self, **kwargs):
        """
        Description: 
        method to change the parameters used for the prediction.

        Inputs:
        - model (str, optional): The name of the model that's going to be used for the prediction.
        There should be a folder with the same name as this on the model's directory that
        contains the corresponding files for this model. If not passed, the model name stays the same.
        - model_dir (str, optional): The absolute path to the directory where the model and language files
        are stored. If they are located where the model is being executed from this parameter can be
        passed a "./". If not passed, the model dir stays the same.
        - language (str, optional): This is the language that will be used for the prediction. On
        the model's directory there should be a folder name "verb-conj" with the verb conjugations for each
        language named "lang-verbs.txt" where "lang" is a short string representing each language: "es" for 
        spanish, "en" for english and so. If not passed, the model language stays the same.

        Outputs:
        None
        """
        for k in kwargs.keys():
            if k == "model":
                self._model = kwargs[k] 
            elif k == "model_dir":
                self._model_dir = kwargs[k]
            elif k == "language":
                self._language = kwargs[k]
            else:
                print(f'{k} is not a valid parameter.')
        self._load_model()

    def train(self, check_model_status = True):
        """
        Description:
        method for carrying out the training on the file define on initialization or with the trainer.new_data method.
        If check_model_status is set to True the algorithm will check for other instances of the model
        both loaded on memory or saved on the models directory. In case it finds any of those, it'll ask
        the user if they want to keep the training going or stop it to make some modifications to avoid overwrite.
        If declared data_format is not "any" it'll check the filetype and stop the training process if 
        these are not the same. Althoug setting it to "any" is easier, it's not recommended since it can
        wrongly include files that are not supposed to act as training files and break the model data.
        After training, it saves the model files inside a folder with the model name in the models directory.

        Inputs:
        - check_model_staus (boolean, optional): if True will check for other instances of the model in memory or
        inside the models directory, and ask the user on wether to keep training with these or not. Default is True.

        Outputs:
        None
        """
        if check_model_status:
            self._model_status()

        detected_format = self._data_path.split(".")[-1]
        if detected_format == self._data_format or self._data_format == "any":
            passing_format = detected_format
        else:
            raise ValueError(f'The detected format "{detected_format}" does not coincide with the passed format "{self._data_format}"')

        self._train_model(self._data_path, passing_format)

        if not self._model_exists():
            os.mkdir(os.path.join(self._model_dir, self._model))
        self._save_model()

    def train_multiple(self, files_dir, check_model_status = True, save_progress = True, remove_done = False):
        """
        Description:
        method for carrying out the training on the files found on files_dir.
        If check_model_status is set to True the algorithm will check for other instances of the model
        both loaded on memory or saved on the models directory. In case it finds any of those, it'll ask
        the user if they want to keep the training going or stop it to make some modifications to avoid overwrite.
        If declared data_format is not "any" it'll check the filetype and skip them from the training process if 
        these are not the same. Althoug setting it to "any" is easier, it's not recommended since it can
        wrongly include files that are not supposed to act as training files and break the model data.

        Inputs:
        - files_dir (str): absolute path to the directory where the training files are located.
        - check_model_staus (boolean, optional): if True will check for other instances of the model in memory or
        inside the models directory, and ask the user on wether to keep training with these or not. Default is True.
        - save_progress (boolean, optional): if True, will save the model files progressively after it's done 
        training each file, and log which files still need to be trained on, in case the algorithm crashes. 
        Otherwise it will only save the progress after all the training progress is finished. Default is True.
        - remove_done (boolean, optional): if True, will remove each file after training on it is done. CAUTION:
        this is not recommended unless a safe copy of the training files is done, in which case is useful to track
        the training progress and continue it in case the ejecution crashes.

        Outputs:
        None
        """
        if check_model_status:
            self._model_status()

        if self._data_format == "any":
            datapath_list = os.listdir(files_dir)
            format_list = [f.split(".")[-1] for f in datapath_list]
        else:
            datapath_list = [f for f in os.listdir(files_dir) if f.split(".")[-1] == self._data_format]
            format_list = [self._data_format]*len(datapath_list)

        for data, form in zip(datapath_list, format_list):
            self._train_model(os.path.join(files_dir, data), form)

            if save_progress:
                if not self._model_exists():
                    os.mkdir(os.path.join(self._model_dir, self._model))
                self._save_model()
                print(f'Saved model in training progress up to {data}.')
                print(f'-> Remaining: {datapath_list[datapath_list.index(data)+1:]}.')

            if remove_done:
                os.remove(os.path.join(files_dir, data))

        if not self._model_exists():
            os.mkdir(os.path.join(self._model_dir, self._model))
        self._save_model()
    
    def new_data(self, data_path, **kwargs):
        """
        Description:
        method for changing the data and/or data format set on initialization.

        Inputs:
        - data_path (str): The absolute path for the file with the phrases that will be used as data.
        - data_format (str, optional): one of "json", "txt" or "any". Ensurance that the file is 
        correctly loaded, the data format of the file will be detected and compared to this parameter. 
        It can be one of two: "json" or "txt". Pass "any" if you don't want any checking of data format.

        Outputs:
        None
        """
        if "data_format" in kwargs.keys():
            self._data_format = kwargs["data_format"]
        self._data_path = data_path
        print("New data loaded.")

    def data_summary(self):
        """
        Description:
        method to generate and log a summary of the dataset characteristics, such as average word count,
        empty sentences; and some examples of how the model will receive this data.

        Inputs:
        None

        Outputs:
        None
        """
        if self._data_format == "txt":
            with open(self._data_path, "r", encoding="utf-8") as f:
                data = f.read().split("\n")
                f.close()
        elif self._data_format == "json":
            with open(self._data_path, "r", encoding="utf-8") as f:
                data_dict = json.load(f)
                data = [d["frase"] for d in data_dict]
                f.close()

        print("-->> Data Summary: <<--")
        print(f'The database has {len(data)} sentences to train with.')
        print(">> These are some of them:")
        for r in [random.randint(0, len(data)-1) for _ in range(5)]:
            print(f'{r}) {data[r]}\n')
        lens = [len(s.split(" ")) for s in data]
        mean = stats.mean(lens)
        std = stats.stdev(lens, mean)
        print('<- - - - ->')
        print(f'Roughly the amount of words per sentence is {int(mean)} +/- {[int(std) if int(std) != 0 else 1][0]}')
        oddball = [i for i in range(len(lens)) if (lens[i] <= mean - 3*std or lens[i] >= mean +3*std)]
        if len(oddball) > 0:
            print(f'{len(oddball)} of them are way off the average amount of words per sentence. This is nothing bad but maybe you should check this is not a bug in the database')
            print(f'These are {oddball}')
        empties = [i for i in range(len(lens)) if (lens[i] == 0 or data[i] == "" or data[i] == " ")]
        if len(empties)>0:
            print('\n<- - - - ->')
            print(f'Some of the phrases are empty. This might raise errors and bugs. We suggest checking them.')
            print(f'These are {empties}')

if __name__ == "__main__":

    tr = trainer("example_database/example.txt", model="en-ex")
    tr.data_summary()
    tr.train()
    


import json
import time
import os
if __name__ == "__main__":
	from calcularPrediccionCache import calcularPrediccionCache
	from generateCacheLanguajeModel import generateCacheLanguajeModel
else:
	from speako.source.calcularPrediccionCache import calcularPrediccionCache
	from speako.source.generateCacheLanguajeModel import generateCacheLanguajeModel


#Script donde se especifica una frase y se obtienen las predicciones
timeStart = time.time() 
#Datos de entrada:
sentence = "Quiero" #Frase inicial sobre la que calcular la predicción
maxNgram = 5 #Máximo orden del N-grama
maxNumPaths = 4 #Máximo número de caminos devueltos tras la ejecución del algoritmo
maxNumWords = 3 #Máximo número de palabras de las frases solución

#Carga de todos los archivos que forman los modelos de lenguaje (normal y de infinitivos)
f = open("listWords.txt", "r")
content = f.read()
listWords = content.split("\n")
f.close()

f = open("listWordsInf.txt", "r")
content = f.read()
listWordsInf = content.split("\n")
f.close()

f = open("dictNgram.txt", "r")
dictNgram = json.load(f)
f.close()
nTot = list(dictNgram.values())[0][3]#Tamaño del diccionario de N-gramas principal

f = open("dictNgramInf.txt", "r")
dictNgramInf = json.load(f)
f.close()

f = open("dictBackoff.txt", "r")
dictBackoff = json.load(f)
f.close()

f = open("dictBackoffInf.txt", "r")
dictBackoffInf = json.load(f)
f.close()

#Carga de formas verbales en español
script_dir = os.path.dirname(__file__)
rel_path = "es-verbs.txt"
fileDictVerbs = os.path.join(script_dir, rel_path) 
verbForms = open(fileDictVerbs, encoding="utf-8") #Se abre el archivo que contiene todas las formas verbales de los verbos en español
verbs_array = verbForms.read().split() #Se obtiene una lista de arrays. Cada array contiene todas las conjugaciones para un mismo verbo.

#Carga de lista de caminos por defecto (a utilizar con aquella frases iniciales que no existan en el diccionario)
f = open("listDefaultPaths.txt", "r", encoding="utf-8")
content = f.read()
listDefaultPaths = content.split("\n")
f.close()

#Parte adicional para diccionarios caché
f = open("listWordsCache.txt", "r")
content = f.read()
listWordsCache = content.split("\n")
f.close()

f = open("listWordsInfCache.txt", "r")
content = f.read()
listWordsCacheInf = content.split("\n")
f.close()

f = open("dictNgramCache.txt", "r")
dictNgramCache = json.load(f)
f.close()
nCache = list(dictNgramCache.values())[0][3]#Tamaño del diccionario de N-gramas de caché

f = open("dictNgramInfCache.txt", "r")
dictNgramCacheInf = json.load(f)
f.close()

f = open("dictBackoffCache.txt", "r")
dictBackoffCache = json.load(f)
f.close()

f = open("dictBackoffInfCache.txt", "r")
dictBackoffCacheInf = json.load(f)
f.close()

time1 = time.time() - timeStart
print("Abrir archivos:",time1)

#Peso, sobre 1, que se le da al diccionario caché
#alfa = (nCache^2)/(nCache+nTot)#Una opción para obtener el parámetro es alfa es ponderarlo en función de los tamaños de los diccionarios
alfa = 0.6 #Otra opción es utilizar valores fijos

#print("frase inicial:",sentence)
#Una vez recopilados todos los datos, se calcula la predicción basada en el algoritmo Viterbi
json_pathsConjugated = calcularPrediccionCache(sentence, maxNgram, maxNumPaths, maxNumWords, listWords, listWordsInf, dictNgram, dictNgramInf, dictBackoff, dictBackoffInf, listWordsCacheInf, dictNgramCache, dictNgramCacheInf, dictBackoffCacheInf, verbs_array, listDefaultPaths, alfa)
#Impresión de los resultados
timeFin = time.time() - timeStart
print(sentence, ": ", json_pathsConjugated,"tiempo:",timeFin)
timeStart2 = time.time()
json_pathsConjugated2 = calcularPrediccionCache(sentence, maxNgram, maxNumPaths, maxNumWords, listWords, listWordsInf, dictNgram, dictNgramInf, dictBackoff, dictBackoffInf, listWordsCacheInf, dictNgramCache, dictNgramCacheInf, dictBackoffCacheInf, verbs_array, listDefaultPaths, alfa)
timeFin2 = time.time() - timeStart2
print(sentence, ": ", json_pathsConjugated2,"tiempo sin apertura:",timeFin2)

try: #Se pregunta al usuario acerca de la opción que desea elegir para completar la frase y, así, guardarla en el diccionario caché.
	choice = int(input("Selecciona la opción para completar la frase: "))
	while(choice>len(json_pathsConjugated) or choice < 1): #Si el usuario indica un valor más alto que las opciones disponibles, se le pregunta hasta que indica un número válido.
		choice = int(input("Valor incorecto. Selecciona una opción válida para completar la frase: "))
except: #Si el usuario no indica un valor numérico, se elige la primera opción por defecto.
	choice = 1
	print("Valor incorrecto. Selecionada opción 1 por defecto.")

timeStartSaving = time.time()
#Guardado de la opción seleccionada por el usuario en los diccionarios caché
fileFormat = "json"
f = open("newSentenceCache.json", "w")
dictSen = dict()
keySen = "frase"
completeSentece =  sentence + " " + json_pathsConjugated.get(str(choice))
dictSen[keySen] = completeSentece
json_data = json.dumps(dictSen) #Se convierte el diccionario en un string JSON
json_paths = json.loads(json_data) # Se crea el objeto JSON
f.write("[")
json.dump(json_paths,f)
f.write("]")
f.close()
file ="newSentenceCache.json"

[listWordsCache, listWordsCacheInf, dictNgramCache, dictNgramCacheInf, dictBackoffCache, dictBackoffCacheInf] = generateCacheLanguajeModel(fileFormat, file, listWordsCache, listWordsCacheInf, dictNgramCache, dictNgramCacheInf, dictBackoffCache, dictBackoffCacheInf)


#Se guardan todos los datos (listas y diccionarios), cada uno en un archivo .txt
f = open("listWordsCache.txt", "w")
for word in listWordsCache:
	if word == listWordsCache[len(listWordsCache)-1]:
		f.write(word)
	else:
		f.write(word+'\n')
f.close()

f = open("listWordsInfCache.txt", "w")
for word in listWordsCacheInf:
    if word == listWordsCacheInf[len(listWordsCacheInf)-1]:
    	f.write(word)
    else:
    	f.write(word+'\n')
f.close()

f = open("dictNgramCache.txt", "w")
json.dump(dictNgramCache, f)
f.close()

f = open("dictNgramInfCache.txt", "w")
json.dump(dictNgramCacheInf, f)
f.close()

f = open("dictBackoffCache.txt", "w")
json.dump(dictBackoffCache, f)
f.close()

f = open("dictBackoffInfCache.txt", "w")
json.dump(dictBackoffCacheInf, f)
f.close()

timeSavingCache = time.time() - timeStartSaving
print(completeSentece,",","tiempo guardado:",timeSavingCache,", alfa:",alfa)


if __name__ == "__main__":
	from generateGeneralLanguajeModel import replaceToInfinitive
	from viterbiNbestCache import viterbiNbestCache
	from viterbiNbestCache import conjugateSentenceCache
	from viterbiNbest import convertToJSON
else:
	from speako.source.generateGeneralLanguajeModel import replaceToInfinitive
	from speako.source.viterbiNbestCache import viterbiNbestCache
	from speako.source.viterbiNbestCache import conjugateSentenceCache
	from speako.source.viterbiNbest import convertToJSON


def calcularPrediccionCache(sentence, maxNgram, maxNumPaths, maxNumWords, listWords, listWordsInf, dictNgram, dictNgramInf, dictBackoff, dictBackoffInf, listWordsCacheInf, dictNgramCache, dictNgramCacheInf, dictBackoffCacheInf, verbs_array, listDefaultPaths, alfa): 
	#Función que realiza la predicción a partir de una frase dada y devuelve los resultados en un array JSON. Utiliza diversas funciones creadas en el archivo 'viterbiNbest'.
	"""Parámetros de entrada:
		- sentence: frase inicial a continuar
		- maxNgram: máximo orden de los N-gramas a evaluar. Por defecto: 5
		- maxNumPaths: máximo número de soluciones a devolver
		- maxNumWords: máximo número de palabras de las frases solución
		- listWords: listado de todas las palabras que aparecen en el diccionario de N-gramas
		- listWordsInf: listado de todas las palabras que aparecen en el diccionario de N-gramas con todos los verbos en infinitivos
		- dictNgram: diccionario de N-gramas
		- dictNgramInf: diccionario de N-gramas de infinitivos
		- dictBackoff: diccionario de bigramas con sus parámetros de backoff
		- dictBackoffInf: diccionario de bigramas de infinitivos con sus parámetros de backoff
		- listWordCachesInf: listado de todas las palabras que aparecen en el diccionario de N-gramas caché con todos los verbos en infinitivos
		- dictNgram: diccionario de N-gramas caché
		- dictNgramCacheInf: diccionario de N-gramas caché de infinitivos
		- dictBackoffCacheInf: diccionario caché de bigramas de infinitivos con sus parámetros de backoff
		- verbs_array: lista con todas las formas verbales de los verbos en español
		- listDefaultPaths: lista de caminos por defecto
		- alfa: peso que se le da al diccionario caché respecto al total, inlcuyendo el diccionario general (valor entre 0 y 1)

		Resultado:
		- json_pathsConjugated: array JSON con todas las predicciones calculadas (ordenadas de mayor a menor probabilidad)
		"""
	sentenceList= sentence.split() #Se convierte la frase inicial en lista de palabras
	sentenceLower = ''
	for w in range(len(sentenceList)): #Se pasan todas las palabras a minúsculas para evitar 'case sensitive'
		sentenceLower = sentenceLower + sentenceList[w].lower() + " "

	sentenceInfList = replaceToInfinitive(sentenceLower.split(), verbs_array) #Se cambian los verbos por su infinitivo
	sentenceInf = ''
	for w in range(len(sentenceInfList)): #Se convierte la lista de palabras con verbos en infinitivo en string
		sentenceInf = sentenceInf + sentenceInfList[w] + " "

	pathsSelected = viterbiNbestCache(sentenceInf, maxNgram, maxNumPaths, maxNumWords, listWordsInf, dictNgramInf, dictBackoffInf, listWordsCacheInf, dictNgramCacheInf, dictBackoffCacheInf, listDefaultPaths, alfa)  #Se ejecuta el algoritmo Viterbi Caché para obtener las predicciones de continuación de la frase inicial

	pathsConjugated = [] #Inicialización de la lista donde se guardarán las frases cuyos verbos estarán conjugados
	for path in pathsSelected: #Conjugación de todos los verbos que aparecen en las frases resultado del algoritmo Viterbi
		pathConj = conjugateSentenceCache(sentenceLower, path, maxNgram, dictNgram, dictNgramCache, verbs_array, alfa)
		pathsConjugated.append(pathConj)

	json_pathsConjugated = convertToJSON(pathsConjugated) #Se convierte la lista de resultados en array JSON

	return json_pathsConjugated #Se devuelve el array JSON


if __name__ == "__main__":
	from generateGeneralLanguajeModel import replaceToInfinitive
	from viterbiNbest import viterbiNbest
	from viterbiNbest import conjugateSentence
	from viterbiNbest import convertToJSON
else:
	from speako.source.generateGeneralLanguajeModel import replaceToInfinitive
	from speako.source.viterbiNbest import viterbiNbest
	from speako.source.viterbiNbest import conjugateSentence
	from speako.source.viterbiNbest import convertToJSON

def calcularPrediccion(sentence, maxNgram, maxNumPaths, listWords, listWordsInf, dictNgram, dictNgramInf, verbs_array, listDefaultPaths): 
	#Función que realiza la predicción a partir de una frase dada y devuelve los resultados en un array JSON. Utiliza diversas funciones creadas en el archivo 'viterbiNbest'.
	"""Parámetros de entrada:
		- sentence: frase inicial a continuar
		- maxNgram: máximo orden de los N-gramas a evaluar. Por defecto: 5
		- maxNumPaths: máximo número de soluciones a devolver
		- listWords: listado de todas las palabras que aparecen en el diccionario de N-gramas
		- listWordsInf: listado de todas las palabras que aparecen en el diccionario de N-gramas con todos los verbos en infinitivos
		- dictNgram: diccionario de N-gramas
		- dictNgramInf: diccionario de N-gramas de infinitivos
		- verbs_array: lista con todas las formas verbales de los verbos en español
		- listDefaultPaths: lista de caminos por defecto

		Resultado:
		- json_pathsConjugated: array JSON con todas las predicciones calculadas (ordenadas de mayor a menor probabilidad)
		"""
	sentenceList= sentence.split() #Se convierte la frase inicial en lista de palabras
	sentenceLower = ''
	for w in range(len(sentenceList)): #Se pasan todas las palabras a minúsculas para evitar 'case sensitive'
		sentenceLower = sentenceLower + sentenceList[w].lower() + " "

	sentenceInfList = replaceToInfinitive(sentenceLower.split(), verbs_array) #Se cambian los verbos por su infinitivo
	sentenceInf = ''
	for w in range(len(sentenceInfList)): #Se convierte la lista de palabras con verbos en infinitivo en string
		sentenceInf = sentenceInf + sentenceInfList[w] + " "

	pathsSelected = viterbiNbest(sentenceInf, maxNgram, maxNumPaths, listWordsInf, dictNgramInf, listDefaultPaths)  #Se ejecuta el algoritmo Viterbi para obtener las predicciones de continuación de la frase inicial

	pathsConjugated = [] #Inicialización de la lista donde se guardarán las frases cuyos verbos estarán conjugados
	for path in pathsSelected: #Conjugación de todos los verbos que aparecen en las frases resultado del algoritmo Viterbi
		pathConj = conjugateSentence(sentenceLower, path, maxNgram, dictNgram, verbs_array)
		pathsConjugated.append(pathConj)

	json_pathsConjugated = convertToJSON(pathsConjugated) #Se convierte la lista de resultados en array JSON

	return json_pathsConjugated #Se devuelve el array JSON


import json
import time
import os
if __name__ == "__main__":
    from calcularPrediccionMultPal import calcularPrediccion
else:
    from speako.source.calcularPrediccionMultPal import calcularPrediccion


#Script donde se especifica una frase y se obtienen las predicciones
timeStart = time.time() 
#Datos de entrada:
sentence = "Los átomos" #Frase inicial sobre la que calcular la predicción
maxNgram = 5 #Máximo orden del N-grama
maxNumPaths = 4 #Máximo número de caminos devueltos tras la ejecución del algoritmo
maxNumWords = 3 #Máximo número de palabras de las frases solución

#Carga de todos los archivos que forman los modelos de lenguaje (normal y de infinitivos)
f = open("listWords.txt", "r")
content = f.read()
listWords = content.split("\n")
f.close()

f = open("listWordsInf.txt", "r")
content = f.read()
listWordsInf = content.split("\n")
f.close()

f = open("dictNgram.txt", "r")
dictNgram = json.load(f)
f.close()

f = open("dictNgramInf.txt", "r")
dictNgramInf = json.load(f)
f.close()

f = open("dictBackoff.txt", "r")
dictBackoff = json.load(f)
f.close()

f = open("dictBackoffInf.txt", "r")
dictBackoffInf = json.load(f)
f.close()

#Carga de formas verbales en español
script_dir = os.path.dirname(__file__)
rel_path = "es-verbs.txt"
fileDictVerbs = os.path.join(script_dir, rel_path) 
verbForms = open(fileDictVerbs, encoding="utf-8") #Se abre el archivo que contiene todas las formas verbales de los verbos en español
verbs_array = verbForms.read().split() #Se obtiene una lista de arrays. Cada array contiene todas las conjugaciones para un mismo verbo.

#Carga de lista de caminos por defecto (a utilizar con aquella frases iniciales que no existan en el diccionario)
f = open("listDefaultPaths.txt", "r", encoding="utf-8")
content = f.read()
listDefaultPaths = content.split("\n")
f.close()

time1 = time.time() - timeStart
print("Abrir archivos:",time1)

#print("frase inicial:",sentence)
#Una vez recopilados todos los datos, se calcula la predicción basada en el algoritmo Viterbi
json_pathsConjugated = calcularPrediccion(sentence, maxNgram, maxNumPaths, maxNumWords, listWords, listWordsInf, dictNgram, dictNgramInf, dictBackoff, dictBackoffInf, verbs_array, listDefaultPaths)
#Impresión de los resultados
timeFin = time.time() - timeStart
print(sentence, ": ", json_pathsConjugated,"tiempo:",timeFin)
timeStart2 = time.time()
json_pathsConjugated2 = calcularPrediccion(sentence, maxNgram, maxNumPaths, maxNumWords, listWords, listWordsInf, dictNgram, dictNgramInf, dictBackoff, dictBackoffInf, verbs_array, listDefaultPaths)
timeFin2 = time.time() - timeStart2
print(sentence, ": ", json_pathsConjugated2,"tiempo sin apertura:",timeFin2)


import json

def generateGeneralLanguajeModel(fileFormat, arg, listWords, listWordsInf, dictNgram, dictNgramInf, dictBackoff, dictBackoffInf, verbs_array): 
	#Función que crea el modelo de lenguaje general de N-gramas (con N<=5) a partir de un archivo .txt o un array JSON
	"""Parámetros de entrada:
		- fileFormat: Formato del archivo a procesar
		- arg: Archivo a procesar
		- listWords: Lista existente de palabras añadidas al diccionario de N-gramas
		- listWordsInf: Lista existente de palabras añadidas al diccionario de N-gramas de infinitivos
		- dictNgram: Diccionario existente de N-gramas
		- dictNgramInf: Diccionario existente de N-gramas de infinitivos
		- dictBackoff: Diccionario existente de backoff por bigramas
		- dictBackoffInf: Diccionario existente de backoff por bigramas de infinitivos
		Resultado:
		- listWords: Lista actualizada de palabras añadidas al diccionario de N-gramas
		- listWordsInf: Lista actualizada de palabras añadidas al diccionario de N-gramas de infinitivos
		- dictNgram: Diccionario actualizado de N-gramas
		- dictNgramInf: Diccionario actualizado de N-gramas de infinitivos
		- dictBackoff: Diccionario actualizado de backoff por bigramas
		- dictBackoffInf: Diccionario actualizado de backoff por bigramas de infinitivos"""

	keySentence = "frase" #Parámetro del array JSON que contiene la frase
	keyWeight = "frecuencia" #Parámetro del array JSON que contiene la frecuencia de ocurrencia de la frase
	valid = True #Indica si el archivo proporcionado es válido para su procesado
	maxNgramSize = 5 #Tamaño máximo de los N-gramas
	finishOperators = ['.', '?','!']  #Carácteres que identifican el final de una frase

	if fileFormat == 'txt': #Caso de archivos .txt que contienen una frase por línea
		input_file = open(arg, encoding="utf-8") #Apertura del archivo .txt con el formato de codificación de caracteres UTF-8, indicado para el uso de caracteres especiales como vocales con tilde o letra Ñ.
		input_file_filtered = replacePunctuation(input_file.read()) #Ver explicación en la función
		text_array = input_file_filtered.split('\n') #Separación del archivo en frases, a partir del caracter \n que indica una nueva línea

		for item in text_array: #Procesado de cada frase para su inclusión en el modelo de lenguaje
			words = item.split() #Se separa la frase en palabras
			wordsInf = replaceToInfinitive(words, verbs_array) #Se obtienen las frases pasando todos los verbos a infinitivo

			for w in words: #Se incluyen las nuevas palabras en la lista de palabras 
				if w.lower() not in listWords:
					listWords.append(w.lower())
			for wI in wordsInf: #Se incluyen las nuevas palabras en la lista de palabras que solo incluye infinitivos
				if wI.lower() not in listWordsInf:
					listWordsInf.append(wI.lower())

			weight = 1 #Para archivos .txt, se cuenta la frecuencia de N-gramas como el conteo de apariciones. Por tanto, cada aparición suma 1.

			sizeSentence = len(words) #Tamaño de la frase en número de palabras
			if(sizeSentence > 0):	#Si la frase no acaba en un signo de puntuación, se le añade uno. En este caso el punto.
				if words[sizeSentence-1] not in finishOperators:
					words.append('.') 
					wordsInf.append('.') 

			for nSize in range(maxNgramSize+1): #Se generan los N-gramas contenidos en la frase que está siendo analizada. Desde el 1-grama (diccionario de palabras) al N-grama.
				dictNgram = addSentenceToDictionary(words, weight, dictNgram, nSize)
				dictNgramInf = addSentenceToDictionary(wordsInf, weight, dictNgramInf, nSize)
			

	elif fileFormat == "json": #Caso de arrays JSON
		input_file = open(arg, encoding="utf-8") #Apertura del array JSON con el formato de codificación de caracteres UTF-8, indicado para el uso de caracteres especiales como vocales con tilde o letra Ñ.
		json_array = json.load(input_file) #Carga del array separado por objetos JSON.

		for item in json_array: 
			words = replacePunctuation(item[keySentence]).split() #Se obtiene la frase del array JSON, se separa en palabras y se procesan los signos de puntuación (ver explicación en función 'replacePunctuation')
			wordsInf = replaceToInfinitive(words, verbs_array) #Se obtienen las frases pasando todos los verbos a infinitivo
			
			for w in words: #Se incluyen las nuevas palabras en la lista de palabras 
				if w.lower() not in listWords:
					#print(w)
					listWords.append(w.lower())
			for wI in wordsInf: #Se incluyen las nuevas palabras en la lista de palabras que solo incluye infinitivos
				if wI.lower() not in listWordsInf:
					listWordsInf.append(wI.lower())

			weight = item[keyWeight] #En este caso, las frases aparecen asociadas a un parámetro de frecuencia. Por tanto, se suma la apariencia de N-gramas en la frase analizada en las mismas unidades que este parámetro.

			sizeSentence = len(words) #Tamaño de la frase en número de palabras
			if(sizeSentence > 0): #Si la frase no acaba en un signo de puntuación, se le añade uno. En este caso el punto.
				if words[sizeSentence-1] not in finishOperators:
					words.append('.')
					wordsInf.append('.')

			for nSize in range(maxNgramSize+1): #Se generan los N-gramas contenidos en la frase que está siendo analizada. Desde el 1-grama (diccionario de palabras) al N-grama.
				dictNgram = addSentenceToDictionary(words, weight, dictNgram, nSize)
				dictNgramInf = addSentenceToDictionary(wordsInf, weight, dictNgramInf, nSize)

	else: #Si el formato de archivo introducido no es válido, se le indica al usuario. Además, se cambia el boolean 'valid' a falso.
		print('Formato no válido. Indique "txt" o "json".')
		valid = False

	if valid: #Se generan los diccionarios del parámetro Backoff tanto para el diccionario regular como para el de infinitivos si se ha proporcionado un archivo de entrada válido
		dictBackoff = generateDictBackoffKatz(dictNgram) 
		dictBackoffInf = generateDictBackoffKatz(dictNgramInf)

	return listWords, listWordsInf, dictNgram, dictNgramInf, dictBackoff, dictBackoffInf #Se devuelven las listas de palabras y diccionarios generados
	

def replacePunctuation(text_array): #Se añade un espacio a los signos de puntuación para asegurar que no forman parte de ninguna palabra, sino que forman una independiente
	"""Parámetros de entrada:
		- text_array: frase inicial
		Resultado:
		- text_array: frase una vez separado los signos de puntuación para formar una nueva palabra"""
	text_array = text_array.replace("¿", "¿ ")
	text_array = text_array.replace("?", " ?")
	text_array = text_array.replace("¡", "¡ ")
	text_array = text_array.replace("!", " !")
	text_array = text_array.replace(".", " . ")
	text_array = text_array.replace(",", " , ")
	text_array = text_array.replace(";", " ; ")
	text_array = text_array.replace(":", " : ")

	return text_array
 
def replaceToInfinitive(words, verbsArray): #Función que reemplaza los verbos por su infinitivo
	"""Parámetros de entrada:
		- words: Lista de palabras
		Resultado:
		- words: Lista de palabras una vez sustituidos los verbos por sus infinitivos"""

	wordsInf = []
	for w in words: #Se comprueba si cada palabra evaluada existe en la lista de conjugaciones verbales
		vb = False
		for verbs in verbsArray:
			single_verbs = verbs.split(',')
			if w.lower() in single_verbs: #Si pertenece a la lista, se obtiene su infinitivo
				vb = True
				wInf = single_verbs[0]
		if vb: #Se añade el infinitivo a la lista de palabras a procesar si es un verbo
			wordsInf.append(wInf)
		else: #Si no, se mantiene la propia palabra
			wordsInf.append(w)

	return wordsInf

def addSentenceToDictionary(words, weight, dictNgram, nSize): #Se añade la frase al modelo de lenguaje general en forma de N-gramas
	"""Parámetros de entrada:
		- words: lista de palabras que contiene la frase
		- weight: peso de la frase a procesar. Para los archivos .txt = 1; para los archivos .json = frecuencia.
		- dictNgram: diccionario de N-gramas
		- nSize: tamaño de los N-gramas 
		Resultado:
		- Diccionario de N-gramas:
			Clave: El N-grama completo
			Valor: lista con los siguientes datos:
				a. Palabra N: Componente N del N-grama
				b. (N-1)-grama: Array de palabras del N-grama desde la 1 hasta la N-1
				c. Número de ocurrencias de la Palabra N tras el (N-1)-grama
				d. Número de ocurrencias del (N-1)-grama
				e. Probabilidad de ocurrencia de la Palabra N respecto al (N-1)-grama: P(N|(N-1)-grama)
			
	Ejemplo de componente en el diccionario:
	N-grama: Vamos a comer arroz
	
	Clave: 'Vamos a comer arroz'
	Valor: ['arroz', ['Vamos', 'a', 'comer'], 2, 8, 0.25]
		a. Palabra N: 'arroz'
		b. (N-1)-grama: ['Vamos', 'a', 'comer']
		c. Ocurrencias de la Palabra N: 2
		d. Ocurrencias del (N-1)-grama: 8
		e. Probabilidad P(N|(N-1)-grama): 25%"""


	listWords = []
	wordsLower = []
	for w in words: #Se ponen todas las palabras en minúsculas antes de añadir al diccionario para evitar 'case sensitive' 
		wordsLower.append(w.lower())

	for i in wordsLower: #Se recorren las palabras contenidas en la frase
		indexList = wordsLower.index(i) #Nueva palabra a procesar
		listWords.append(i)  #Se añade la palabra a la lista de palabras que han aparecido anteriormente en la frase
		listDict = []
		if indexList >= (nSize-1): #Para recoger un N-grama, la frase debe conetener, al menos, las mismas palabras que el tamaño del mismo
			keyDict = ""
			for k in range(nSize): #La clave del diccionario será el propio N-grama, que es único. Está formado por la palabra y las N-1 anteriores.
				keyDict = wordsLower[indexList-k] + " " + keyDict
			listDict.append(i) #Se añade el primer parámetro de la lista de valores (Palabra N)
			listPrev = wordsLower[(indexList-(nSize-1)):indexList:1]
			listDict.append(listPrev) #Se añade el segundo parámetro de la lista de valores: (N-1)-grama

			[num, numTot, per, dictNgram] = updateDict(listPrev, keyDict, weight, dictNgram) #Se actualiza el diccionario tras el procesado del nuevo N-grama

			listDict.append(num) #Se añade el tercer parámetro de la lista de valores: ocurrencias de la Palabra N tras el (N-1)-grama
			listDict.append(numTot) #Se añade el cuarto parámetro de la lista de valores: ocurrencias del (N-1)-grama
			listDict.append(per) #Se añade el quinto parámetro de la lista de valores: probabilidad P(N|(N-1)-grama)
			dictNgram[keyDict] = listDict #Se adjunta la lista de valores a la clave del diccionario

	return dictNgram #Se devuelve el diccionario actualizado

def updateDict(listPrev, keyDict, weight, dictNgram): #Función que actualiza el diccionario tras el procesado de un nuevo N-grama
	"""Parámetros de entrada:
		- listPrev: lista de palabras que componen el (N-1)-grama a procesar
		- keyDict: clave del diccionario correspondiente al N-grama a procesar (el propio N-grama)
		- weight: peso del nuevo N-grama 
		- dictNgram: diccionario de N-gramas

		Resultado:
		- num: número de ocurrencias totales de la palabra N tras el (N-1)-grama tras el procesado del nuevo N-grama
		- numTot: número de ocurrencias totales del (N-1)-grama tras el procesado del nuevo N-grama
		- per: porcentaje actualizado 'num'/'numTot', equivalente a P(N|(N-1)-grama)
		- dictNgram: diccionario actualizado
		"""

	keysDictExt = list(dictNgram.keys()) #Se obtiene la lista de claves del diccionario
	if len(keysDictExt) > 0: #Se comprueba que el diccionario no está vacío
		valuesDictExt = list(dictNgram.values()) #Se obtiene la lista de valores del diccionario
		valuesDictExtTras = list(zip(*valuesDictExt)) #Se traspone la lista de valores del diccionario para obtener cada uno de los datos
		valuesPrev = valuesDictExtTras[1] #Obtención de los (N-1)-grama en forma de lista de palabras
		numTotPrev = valuesDictExtTras[3] #Obtención del dato 'numTot' o número de ocurrencias de cada (N-1)-grama
		if listPrev in valuesPrev: #Caso de que el (N-1)-grama exista
			indexPrev = valuesPrev.index(listPrev) #Se obtiene el índice donde se encuentra el (N-1)-grama
			numTot = numTotPrev[indexPrev] + weight #Se actualizad el parámetro 'numTot'  del (N-1)-grama
			for indexEv in range(len(keysDictExt)):
				if(listPrev == valuesPrev[indexEv]): #Se actualizan los parámetros 'numTot' y 'per' de todos los N-gramas del diccionario que comparten el mismo (N-1)-grama debido al incrementeo del 'numTot'
					keyChange = keysDictExt[indexEv] 
					numPrev = dictNgram[keyChange][2]
					dictNgram[keyChange][3] = numTot
					dictNgram[keyChange][4] = numPrev / numTot
		else: #Si el (N-1)-grama no existe, se inicializa
			numTot = weight

		if keyDict in dictNgram: #Si el N-grama ya existía, se actualiza su ocurrencia 'num'
			num = dictNgram[keyDict][2] + weight
		else: #Si el N-grama no existía, se inicializa su ocurrencia 'num'
			num = weight
	else: #El primer N-grama se añade al diccionario siendo 'num' y 'numTot' el peso del N-grama 
		num = weight
		numTot = weight

	per = num / numTot #Cálculo de la probabilidad de ocurrencia N|(N-1)-grama

	return num, numTot, per, dictNgram #Se devuelven los valores 'num', 'numTot' y 'per' del N-grama procesado además del diccionario actualizado

def generateDictBackoffKatz(dictNgram): #Función que genera el diccionario de parámetros backoff asociados a los bigramas incluídos en el diccionario de N-gramas proporcionado
	"""Parámetros de entrada:
		- dictNgram: diccionario de N-gramas

		Resultado:
		- dictBackoff: diccionario de bigramas con sus parámetros de backoff
			Clave: El bigrama completo
			Valor: lista con los siguientes datos:
				a. Palabra 'i': Componente 'i' del bigrama o segunda palabra
				b. Palabra 'i-1': Componente 'i-1' del bigrama o primera palabra
				c. Parámetro 'c_katz' del bigrama analizado
				d. #umatorio de los 'c_katz' que comparten la primera palabra, 'w_{i-1}', del bigrama analizado
				e. #Probabilidad de backoff del bigrama analizado: P_backoff(wi_{i-1}) = p_katz(w_i|w_{i-1})
			"""
	
	dictBackoff = {} #Inicialización del diccionario de parámetro backoff
	keysDictExt = list(dictNgram.keys())  #Se extraen las claves del diccionario de N-gramas

	for nGram in keysDictExt: #Se recorre el diccionario de N-gramas en busca de bigramas con el objetivo de calcular su parámetro de backoff y añadirlos a este diccionario
		listWords = nGram.split()
		if len(listWords)==2:
			dictBackoff = addBigramToDictBackoffKatz(nGram, dictNgram, dictBackoff)
			
	return dictBackoff #Se devuelve el diccionario de bigramas con sus parámetros de backoff actualizado


def addBigramToDictBackoffKatz(bigram, dictNgram, dictBackoff): #Función que calcula el parámetro de backoff para un bigrama según el método de Katz
	"""Parámetros de entrada:
			- bigram: bigrama sobre el que calcular su backoff y añadir al diccionario
			- dictNgram: diccionario de N-gramas
			- dictBacoff: diccionario de bigramas con sus parámetros de backoff

			Resultado:
			- dictBackoff: diccionario actualizado de bigramas con sus parámetros de backoff 
				Clave: El bigrama completo
				Valor: lista con los siguientes datos:
					a. Palabra 'i': Componente 'i' del bigrama o segunda palabra
					b. Palabra 'i-1': Componente 'i-1' del bigrama o primera palabra
					c. Parámetro 'c_katz' del bigrama analizado
					d. #umatorio de los 'c_katz' que comparten la primera palabra, 'w_{i-1}', del bigrama analizado
					e. #Probabilidad de backoff del bigrama analizado: P_backoff(wi_{i-1}) = p_katz(w_i|w_{i-1})
			"""
	k = 5 #Katz recomienda utilizar el parámetro 'k' con valor 5
	#Inicialización de parámetros de la función
	nr = 0 
	nr1 = 0
	n1 = 0
	nk1 = 0

	listWords = bigram.split() #Se separa el bigrama en 'w0' ('w_i') y 'w1' ('w_{i-1}')
	w0 = listWords[1]
	w1 = listWords[0] 

	keysDictExt = list(dictNgram.keys())  #Se extraen las claves del diccionario de N-gramas
	if len(keysDictExt) > 0: #Si el diccionario de N-gramas tiene valores, se obtienen (explicados en la función 'addSentenceToDictionary')
		valuesDictExt = list(dictNgram.values()) 
		valuesDictExtTras = list(zip(*valuesDictExt)) 
		values = valuesDictExtTras[0]
		valuesPrev = valuesDictExtTras[1]
		valuesNum = valuesDictExtTras[2]
		valuesNumTot = valuesDictExtTras[3]
		prob = valuesDictExtTras[4]

	ind = keysDictExt.index(bigram) #Se obtiene el índice del diccionario de N-gramas donde se encuentra el bigrama analizado
	r = valuesNum[ind] #Se obtiene el número de ocurrencias del bigrama analizado
	for i in range(len(valuesNum)): #Se calculan los parámetros del método de Katz:
		if valuesNum[i] == r: #Número de N-gramas en el diccionario que ocurren las mismas veces, 'r', que el bigrama analizado
			nr = nr + 1
		elif valuesNum[i] == (r+1): #Número de N-gramas en el diccionario que ocurren una vez más, 'r+1', que el bigrama analizado
			nr1 = nr1 + 1 

		if valuesNum[i] == (k+1): #Número de N-gramas en el diccionario que ocurren una vez más, 'k+1', que el parámetro 'k' definido al inicio
			nk1 = nk1 + 1
		if valuesNum[i] == 1: #Número de N-gramas en el diccionario que ocurren una vez
			n1 = n1 + 1

	r1= (r+1)*nr1/nr #Cálculo del parámetro r*
	#Cálculo del parámetro dr
	if n1==0:
		dr = 0
	else:
		numDr = r1/r - (k+1)*nk1/n1 
		denDr = 1 - (k+1)*nk1/n1
		dr = numDr/denDr 

	cKatz = dr*r #Cálculo del valor 'c_katz(wi_{i-1})'
	cKatzTotal = getCKatzTotal(w1, cKatz, dictBackoff) #Obtención del sumatorio de todos los 'c_katz(wi_{i-1})' que comparten el mismo 'w_{i-1}' o primera palabra del bigrama
	dictBackoff = updateCKatzTotal(w1, cKatzTotal, dictBackoff) #Se actualiza el diccionario de backoff tras el procesado del nuevo bigrama

	#Se calcula la probabilidad de backoff, 
	if cKatzTotal == 0: #En caso de que el denominador sea 0, esta probabilidad es cero
		probBackoff = 0 
	else:  #En caso contrario, se calcula la probabilidad de backoff como 'P_backoff(wi_{i-1}) = p_katz(w_i|w_{i-1}) = c_katz(wi_{i-1})/SUM_{w_i}[c_katz(wi_{i-1})]' 
		probBackoff = cKatz/cKatzTotal #O, explicado de otro modo, dividiendo el parámetro 'c_katz' del bigrama analizado entre el sumatorio de los 'c_katz' de todos los bigramas que comparten su primera palabra ('w_{i-1}')

	keyDict = bigram #Se añade el bigrama analizado al diccionario de backoff; su clave es el propio bigrama y sus valores los siguientes:
	listDict = []
	listDict.append(w0) #Segunda palabra del bigrama: 'w_i'
	listDict.append(w1) #Primera palabra del bigrama: 'w_{i-1}'
	listDict.append(cKatz) #Parámetro 'c_katz' del bigrama analizado
	listDict.append(cKatzTotal) #Sumatorio de los 'c_katz' que comparten la primera palabra, 'w_{i-1}', del bigrama analizado
	listDict.append(probBackoff) #Probabilidad de backoff del bigrama analizado
	dictBackoff[keyDict] = listDict

	return dictBackoff #Se devuelve el diccionario de backoff actualizado


def getCKatzTotal(w1, cKatz, dictBackoff): #Función que obtiene el sumatorio de todos los 'c_katz(wi_{i-1})' que comparten la misma palabra w_{i-1}
	"""Parámetros de entrada:
			- w1: Primera palabra del bigrama analizado,'w_{i-1}'
			- cKatz: Valor 'c_katz(wi_{i-1})' del bigrama analizado
			- dictBackoff: diccionario de bigramas con sus parámetros de backoff

			Resultado:
			- cKatzTotal: sumatorio de todos los 'c_katz(wi_{i-1})' que comparten la misma palabra 'w_{i-1}'
			"""
	keysDictBO = list(dictBackoff.keys()) #Se extraen las claves del diccionario de backoff
	if len(keysDictBO) > 0: #Si el diccionario de backoff tiene valores, se obtienen (explicados en las funciones 'generateDictBackoffKatz' y 'addBigramToDictBackoffKatz')
		valuesDictBO = list(dictBackoff.values()) 
		valuesDictBOTras = list(zip(*valuesDictBO))
		valuesBO = valuesDictBOTras[0]
		valuesPrevBO = valuesDictBOTras[1]
		valuesCK = valuesDictBOTras[2]
		valuesCKTot = valuesDictBOTras[3]
		probK = valuesDictBOTras[4]
	
		if w1 in valuesPrevBO: #Si la palabra 'w1' ('w_{i-1}') ya ha sido utilizado en otro bigrama incluido en el diccionario, se obtiene el valor actual de 'cKatzTotal'
			ind = valuesPrevBO.index(w1)
			cKatzTotal = valuesCKTot[ind]
		else: #Si no, se inicializa a cero
			cKatzTotal = 0

		cKatzTotal = cKatzTotal + cKatz #Se añade el valor 'cKatz' del bigrama analizado a 'cKatzTotal'

	else: #Si el diccionario está vacío, el valor de 'cKatzTotal' será el valor de 'cKatz' del bigrama bajo análisis ya que será su primer valor
		cKatzTotal = cKatz

	return cKatzTotal #Se devuelve el valor 'cKatzTotal' de los bigramas que comparten la misma palabra 'w1' ('w_{i-1}') que el bigrama analizado

def updateCKatzTotal(w1, cKatzTotal, dictBackoff): 
	"""Se actualiza el diccionario de backoff tras el procesado de un nuevo bigrama.
		Se deben actualizar los valores 'cKatzTotal' y 'probBackoff' de todos los bigramas que compartan la misma primera palabra, 'w1' ('w_{i-1}'), que el bigrama procesado
		Parámetros de entrada:
			- w1: primera palabra, 'w1' ('w_{i-1}'), del bigrama analizado
			- cKatzTotal: valor 'cKatzTotal' de los bigramas que comparten la misma palabra 'w1' ('w_{i-1}')
			- dictBackoff: diccionario de bigramas con sus parámetros de backoff

			Resultado:
			- dictBackoff:  diccionario actualizado de bigramas con sus parámetros de backoff
			"""
	keysDictBO = list(dictBackoff.keys()) #Se extraen las claves del diccionario de backoff
	if len(keysDictBO) > 0: #Si el diccionario de backoff tiene valores, se obtienen (explicados en las funciones 'generateDictBackoffKatz' y 'addBigramToDictBackoffKatz')
		valuesDictBO = list(dictBackoff.values()) 
		valuesDictBOTras = list(zip(*valuesDictBO)) 
		valuesBO = valuesDictBOTras[0]
		valuesPrevBO = valuesDictBOTras[1]
		valuesCK = valuesDictBOTras[2]
		valuesCKTot = valuesDictBOTras[3]
		probK = valuesDictBOTras[4]

		for i in range(len(valuesPrevBO)): #Se evalúa el diccionario de backoff en busca de los bigramas que compartan la misma primera palabra, 'w1' ('w_{i-1}')
			if w1 == valuesPrevBO[i]:
				cKatzAct = valuesCK[i] #Se obtiene el valor 'cKatz' del bigrama encontrado
				keyChange = keysDictBO[i]
				dictBackoff[keyChange][3] = cKatzTotal #Se guarda el nuevo valor de 'cKatzTotal'
				if cKatzTotal == 0: #Se asegura que el denominador no sea 0
					probAct = 0
				else:
					probAct = cKatzAct / cKatzTotal #Se calcula la nueva probabilidad de backoff del bigrama encontrado
				dictBackoff[keyChange][4] = probAct #Se guarda el valor actualizado de la probabilidad de backoff del bigrama encontrado

	return dictBackoff #Se devuelve el diccionario de backoff actualizado


import json
import time
if __name__ == "__main__":
	from generateCacheLanguajeModel import generateCacheLanguajeModel
else:
	from speako.source.generateCacheLanguajeModel import generateCacheLanguajeModel


#Script que actualiza los modelos de lenguaje caché tras la inclusión de una nueva frase

#Se especifican los archivos con los que crear los modelos de lenguaje en dos listas: fileFormat (formato de los archivos) y file (nueva frase)
#Como ejemplo, se propone un frase en formato .json

timeStart = time.time()
fileFormat = "json"
file ="newSentenceCache.json" #Archivo que contiene la frase de ejemplo

#Luego, se cargan las listas y diccionarios a actualizar. Son los siguientes:

f = open("listWordsCache.txt", "r")
content = f.read()
listWords = content.split("\n")
f.close()

f = open("listWordsInfCache.txt", "r")
content = f.read()
listWordsInf = content.split("\n")
f.close()

f = open("dictNgramCache.txt", "r")
dictNgram = json.load(f)
f.close()

f = open("dictNgramInfCache.txt", "r")
dictNgramInf = json.load(f)
f.close()

f = open("dictBackoffCache.txt", "r")
dictBackoff = json.load(f)
f.close()

f = open("dictBackoffInfCache.txt", "r")
dictBackoffInf = json.load(f)
f.close()

rel_path = "verb-conj/es-verbs.txt"
verbForms = open(rel_path, encoding="utf-8") #Se abre el archivo que contiene todas las formas verbales de los verbos en español
verbs_array = verbForms.read().split()

#Se añade una nueva frase al diccionario de caché
[listWords, listWordsInf, dictNgram, dictNgramInf, dictBackoff, dictBackoffInf] = generateCacheLanguajeModel(fileFormat, file, listWords, listWordsInf, dictNgram, dictNgramInf, dictBackoff, dictBackoffInf, verbs_array)

#Se guardan todos los datos (listas y diccionarios), cada uno en un archivo .txt
f = open("listWordsCache.txt", "w")
for word in listWords:
	if word == listWords[len(listWords)-1]:
		f.write(word)
	else:
		f.write(word+'\n')
f.close()

f = open("listWordsInfCache.txt", "w")
for word in listWordsInf:
    if word == listWordsInf[len(listWordsInf)-1]:
    	f.write(word)
    else:
    	f.write(word+'\n')
f.close()

f = open("dictNgramCache.txt", "w")
json.dump(dictNgram, f)
f.close()

f = open("dictNgramInfCache.txt", "w")
json.dump(dictNgramInf, f)
f.close()

f = open("dictBackoffCache.txt", "w")
json.dump(dictBackoff, f)
f.close()

f = open("dictBackoffInfCache.txt", "w")
json.dump(dictBackoffInf, f)
f.close()

#Impresión de los resultados
timeFin = time.time() - timeStart
print("tiempo:",timeFin)


from distutils.core import setup

setup(
    name = 'speako',
    version = '0.1.0',
    author = 'OTTAA Dev Team',
    author_email = 'info@ottaaproject.com',
    packages = ['speako', 'speako.source'],
    scripts = ['ipy-nbs/basic_use.ipynb',
                'ipy-nbs/interactive_prediction.ipynb',
                'ipy-nbs/mult_train.ipynb',
                'ipy-nbs/saving_cache.ipynb',               
                ],
    data_files = ['speako/lang/es-verbs.txt',
                  'speako/lang/en-verbs.txt',
                  ],
    url = 'https://ottaa-project.github.io/PictogramsPredictionsLibrary/',
    license = '',
    description = 'package for language prediction used by OTTAA apps',
    long_description = open('README.txt').read(),
    install_requires=[],
)    


import json
import math

def viterbiNbest(sentence, maxNgram, maxNumPaths, listWords, dictNgram, dictBackoff, listDefaultPaths): #Función que devuelve las listas de palabras más probables para continuar la frase inicial dada
	"""Parámetros de entrada:
		- sentence: frase inicial a continuar
		- maxNgram: máximo orden de los N-gramas a evaluar. Por defecto: 5
		- maxNumPaths: máximo número de soluciones a devolver
		- listWords: listado de todas las palabras que aparecen en el diccionario de N-gramas
		- dictNgram: diccionario de N-gramas
		- dictBackoff: diccionario de bigramas con sus parámetros de backoff
		- listDefaultPaths: lista de caminos por defecto

		Resultado:
		- pathsSelected: caminos más probables (en forma de lista de strings) para continuar la frase inicial
		"""
	listSign = [".", "?", "!", "¿", "¡", ",",";"] #Se indica la lista de signos de puntuación para no considerarlos en la evaluación de palabras
	sentenceWords = sentence.split() #Se transforma la frase inicial en lista de palabras
	nIni = min(len(sentenceWords), maxNgram-1) #Se obtiene la longitud de la frase inicial (como máximo el orden máximo del N-grama establecido menos 1 ya que hay que añadir la nueva palabra a evaluar)
	nGram = [] #Se inicializa la lista de palabras que forman un N-grama
	pathsSaved = dict() #Se inicializa el diccionario de caminos guardados
	listWordsEv = list() #Lista de palabras a evaluar
	nGramBack = list() #Lista de palabras de la parte inicial de la frase para el cálculo del backoff

	probPaths = list() #Lista de las probabilidades de los caminos activos

	keysDict = list(dictNgram.keys()) #Se obtienen las claves del diccionario de N-gramas en forma de lista
	valuesDict= list(dictNgram.values()) #Se obtienen los valores del diccionario de N-gramas en forma de lista

	keysDictBack = list(dictBackoff.keys()) #Se obtienen las claves del diccionario de Backoff en forma de lista
	valuesDictBack = list(dictBackoff.values()) #Se obtienen los valores del diccionario de Backoff en forma de lista

	while len(listWordsEv) == 0 and len(sentenceWords)>0: #Se buscan palabras objetivo mientra no haya ninguna y el N-grama a evaluar no esté vacío
		for key, value in dictNgram.items(): #Se obtienen aquellas palabras que se encuentran en las frases del diccionario que incluyen la frase inicial
			if sentence in key:
				wordsSentece = key.split()
				for w in wordsSentece:
					if w not in listWordsEv and w not in listSign: #Se guardan aquellas palabras que aún no se encuentran en la lista, no se considerando los signos de puntuación
						listWordsEv.append(w)
		if len(listWordsEv) == 0 and len(sentenceWords)>0: #Si la frase completa no se encuentra en el diccionario, se va acortando el N-grama eliminando las palabras iniciales
			nGramBack.append(sentenceWords[0])
			sentenceWords.remove(sentenceWords[0])
			sentence = ' '.join(sentenceWords)


	n = min(len(sentenceWords), maxNgram-1) #Se obtiene el número de palabras iniciales sobre las que trabajar (como máximo el orden máximo del N-grama establecido menos 1 ya que hay que añadir la nueva palabra a evaluar)
	nBack = nIni - n #Diferencia entre el número de palabras de la frase inicial y el de la frase con la que trabajar; utilizado para el cálculo del backoff
	probBack = 1 #Se inicializa la probabilidad de backoff a 1; no influye en aquellos casos no necesarios
	if nBack > 0  and len(sentenceWords) > 0:
		nGramBack.append(sentenceWords[0]) #Se añade la primera palabra de la parte final de la frase (evaluación diccionario) para incluirla en backoff
		try: #Se evaluan los bigrama de inicio de frase para recoger su parámetro de backoff
			for b in range(nBack): 
				bigram = nGramBack[b]+" "+nGramBack[b+1]
				indexSentenceBack = keysDictBack.index(bigram)
				probBack = probBack*valuesDictBack[indexSentenceBack][4]
		except: #Si algún bigrama no se encuentra en el diccionario, se inicializa la probabilidad de backoff
			probBack = 1

	try:
		for i in range(n):
			listWordsEv.remove(sentenceWords[len(sentenceWords)-(i+1)]) #Se elimina las palabras de la frase inicial para no repetirlas en la predicción
	except:
		listWordsEv

	for i in range(n): #Se construye el N-grama inicial sobre el que trabajar
		nGram.append(sentenceWords[len(sentenceWords)-n+i])

	for word in listWordsEv: #Se recorre la lista de palabras para calcular su probabilidad y se añade a la lista de probabilidades
		prob = evaluateNgram(word, nGram, keysDict, valuesDict, nBack, probBack)
		pathsSaved[word] = prob
		#print(word,prob)

	pathsSavedOrdered = dict(sorted(pathsSaved.items(), key=lambda item: item[1], reverse = True)) #Se ordenan las palabras evaluadas en orden decreciente de probabilidad
	pathsSO = list(pathsSavedOrdered.keys()) #Se obtiene la lista de caminos guardados ordenada por probabilidad

	pathsSelected = list() 
	for pthSd in range(min(maxNumPaths,len(pathsSO))): #Se recogen los K caminos más probables, siendo K el número máximo de caminos a devolver
		pathsSelected.append(pathsSO[pthSd])

	ind = 0
	while len(pathsSelected) < maxNumPaths:
		pathsSelected.append(listDefaultPaths[ind])
		ind = ind + 1

	return pathsSelected #Se devuelven los caminos más probables


def evaluateNgram(word, nGram, keysDict, valuesDict, nBack, probBack): #Función que calcula la probabilidad logarítmica P(w|N-grama)
	"""Parámetros de entrada:
		- word: palabra a añadir tras el N-grama proporcionado (w)
		- nGram: N-grama inicial
		- keysDict: lista de claves del diccionario de N-gramas
		- valuesDict: lista de valores del diccionario de N-gramas
		- nBack: Palabras iniciales de la frase a evaluar con las que calcular el backoff
		- probBack: Probabilidad de backoff

		Resultado:
		- logprob: probabilidad logarítmica P(N|(N-1)-grama)
		"""
	sentence = '' #Inicialización de la frase a evaluar
	nGramEv = []

	for w in nGram: #Se añade la nueva palabra al N-grama inicial en una nueva lista
		nGramEv.append(w)
	
	nGramEv.append(word)
	for i in range(len(nGramEv)): #Se convierte el N-grama inicial, dado en forma de lista, en un string
		sentence = sentence + nGramEv[i] + " "

	try: #Se obtiene la probabilidad de la frase evaluada
		indexSentence = keysDict.index(sentence)
		prob = valuesDict[indexSentence][4]
	except:
		prob = 0

	prob = prob * probBack

	if prob>0: #Se calcula la probabilidad logarímitica; si la probabilidad fuera 0, se asigna un valor muy bajo (realmente: -inf)
		logprob = math.log(prob) 
	else:
		logprob = -10000

	return logprob #Se devuelve la probabilidad logarítmica

def conjugateSentence(sentence, path, maxNgram, dictNgram, verbs_array):#Función que sustituye los verbos (en infinitivo) de una frase dada por su conjugación más probable
	"""Parámetros de entrada:
		- sentence: frase inicial
		- path: frase final a procesar
		- maxNgram: máximo orden de los N-gramas a evaluar. Por defecto: 5
		- dictNgram: diccionario de N-gramas
		- verbForms: archivo que contiene todas las formas verbales de los verbos en español

		Resultado:
		- final_sentence: frase conjugada
		"""
	wordsSentece = sentence.split() #Se convierte la frase inicial en lista de palabras
	lenIni = len(wordsSentece) #Se obtiene el número de palabras de la frase inicial
	final_path = wordsSentece #Inicialización de la lista de palabras que formarán la frase final

	keysDict = list(dictNgram.keys()) #Se obtienen las claves del diccionario de N-gramas en forma de lista
	valuesDict= list(dictNgram.values()) #Se obtienen los valores del diccionario de N-gramas en forma de lista

	infinitives_array = [] #Se inicializa la lista de infinitivos
	for verbs in verbs_array: #Se recorren todos los verbos para adjuntar a la lista sus infinitivos (primera posición del array donde aparecen todas las conjugaciones para cada verbo)
		single_verbs = verbs.split(',')
		infinitives_array.append(single_verbs[0].lower())

	sentenceWords = path.split() #Se convierte la frase inicial en una lista de palabras
	for word in sentenceWords: #Se comprueba si cada palabra es un infinitivo
		if word in infinitives_array: #Si lo es, se intercambia por su conjugación más probable
			indexVerb = infinitives_array.index(word) #Primero, se identifica el verbo encontrado
			conj = verbs_array[indexVerb].split(',') #Para recoger todas sus formas verbales en una lista
			nGram = final_path #Luego, se forma el N-grama anterior al verbo encontrado. 
			while len(nGram) > maxNgram: #Si es mayor que la longitud máxima, se van eliminando las palabras iniciales hasta ajustar su longitud
				nGram.pop(0)

			maxProb = -10000
			nIni = 0
			while maxProb == -10000 and nIni <= len(nGram):#Se evaluan las conjugaciones objetivo mientras no se haya encontrado ninguna palabra y el N-grama a evaluar no esté vacío
				probList = []
				for verb in conj: #Posteriormente, se evalúa cada forma verbal tras el N-grama anterior
					prob = evaluateNgram(verb, nGram[nIni:len(nGram)], keysDict, valuesDict, 0, 1)
					probList.append(prob)
				maxProb = max(probList) #Se calcula la máxima probabilidad de entre todas las formas verbales evaluadas
				if maxProb == -10000 :#Si el N-grama completo no encuentra conjugaciones, se va acortando el N-grama eliminando las palabras iniciales
					nIni = nIni + 1
	
			indexMax = probList.index(maxProb) #Y se obtiene el índice donde ha acontecido esta máxima probabilidad 

			wordConj = conj[indexMax] #Finalmente, se recoge esta forma verbal
			final_path.append(wordConj) #Y se añade a la lista final de palabras
		else: #Si no, se añade a la lista final tal cual
			final_path.append(word)


	for i in range(lenIni): #Se eliminan las palabras de la frase inicial del resultado
		try:
			final_path.pop(0)
		except:
			final_path

	final_sentence = " ".join(final_path) #Se convierte la lista de palabras en un string que conforma la frase final

	return final_sentence #Se devuelve la frase con los verbos conjugados

def convertToJSON(paths):#Función que convierte en objeto JSON una lista de strings
	"""Parámetros de entrada:
		- paths: lista de strings

		Resultado:
		- json_paths: objeto json cuya clave es el número de orden y su valor el string
		"""
	dictPaths = dict() #Se crea un diccionario como paso previo a la creación del objeto JSON
	ind = 1 #Se inicializa el índice que servirá como clave del diccionario

	for p in paths: #Se recorre la lista de strings para asignarlas como valor al diccionario cuya clave es el orden o posición en la lista
		dictPaths[ind] = p
		ind = ind + 1
 
	json_data = json.dumps(dictPaths) #Se convierte el diccionario en un string JSON
	json_paths = json.loads(json_data) # Se crea el objeto JSON
 
	return json_paths #Se devuelve el objeto JSON creado 



import json
if __name__ == "__main__":
	from generateGeneralLanguajeModel import addSentenceToDictionary
	from generateGeneralLanguajeModel import replacePunctuation
	from generateGeneralLanguajeModel import replaceToInfinitive
	from generateGeneralLanguajeModel import generateDictBackoffKatz
else:
	from speako.source.generateGeneralLanguajeModel import addSentenceToDictionary
	from speako.source.generateGeneralLanguajeModel import replacePunctuation
	from speako.source.generateGeneralLanguajeModel import replaceToInfinitive
	from speako.source.generateGeneralLanguajeModel import generateDictBackoffKatz

def generateCacheLanguajeModel(fileFormat, arg, listWords, listWordsInf, dictNgram, dictNgramInf, dictBackoff, dictBackoffInf, verbs_array): 
	#Función que crea el modelo de lenguaje general de N-gramas (con N<=5) a partir de un archivo .txt o un array JSON
	"""Parámetros de entrada:
		- fileFormat: Formato del archivo a procesar
		- arg: Archivo a procesar
		- listWords: Lista existente de palabras añadidas al diccionario de N-gramas
		- listWordsInf: Lista existente de palabras añadidas al diccionario de N-gramas de infinitivos
		- dictNgram: Diccionario existente de N-gramas
		- dictNgramInf: Diccionario existente de N-gramas de infinitivos
		- dictBackoff: Diccionario existente de backoff por bigramas
		- dictBackoffInf: Diccionario existente de backoff por bigramas de infinitivos
		Resultado:
		- listWords: Lista actualizada de palabras añadidas al diccionario de N-gramas
		- listWordsInf: Lista actualizada de palabras añadidas al diccionario de N-gramas de infinitivos
		- dictNgram: Diccionario actualizado de N-gramas
		- dictNgramInf: Diccionario actualizado de N-gramas de infinitivos
		- dictBackoff: Diccionario actualizado de backoff por bigramas
		- dictBackoffInf: Diccionario actualizado de backoff por bigramas de infinitivos"""

	keySentence = "frase" #Parámetro del array JSON que contiene la frase
	keyWeight = "frecuencia" #Parámetro del array JSON que contiene la frecuencia de ocurrencia de la frase
	valid = True #Indica si el archivo proporcionado es válido para su procesado
	maxNgramSize = 5 #Tamaño máximo de los N-gramas
	finishOperators = ['.', '?','!']  #Carácteres que identifican el final de una frase

	if fileFormat == 'txt': #Caso de archivos .txt que contienen una frase por línea
		input_file = open(arg, encoding="utf-8") #Apertura del archivo .txt con el formato de codificación de caracteres UTF-8, indicado para el uso de caracteres especiales como vocales con tilde o letra Ñ.
		input_file_filtered = replacePunctuation(input_file.read()) #Ver explicación en la función
		text_array = input_file_filtered.split('\n') #Separación del archivo en frases, a partir del caracter \n que indica una nueva línea

		for item in text_array: #Procesado de cada frase para su inclusión en el modelo de lenguaje
			words = item.split() #Se separa la frase en palabras
			wordsInf = replaceToInfinitive(words, verbs_array) #Se obtienen las frases pasando todos los verbos a infinitivo

			for w in words: #Se incluyen las nuevas palabras en la lista de palabras 
				if w.lower() not in listWords:
					listWords.append(w.lower())
			for wI in wordsInf: #Se incluyen las nuevas palabras en la lista de palabras que solo incluye infinitivos
				if wI.lower() not in listWordsInf:
					listWordsInf.append(wI.lower())

			weight = 1 #Cada nueva frase se suma con peso 1 a los diccionarios existentes.

			sizeSentence = len(words) #Tamaño de la frase en número de palabras
			if(sizeSentence > 0):	#Si la frase no acaba en un signo de puntuación, se le añade uno. En este caso el punto.
				if words[sizeSentence-1] not in finishOperators:
					words.append('.') 
					wordsInf.append('.') 

			for nSize in range(maxNgramSize+1): #Se generan los N-gramas contenidos en la frase que está siendo analizada. Desde el 1-grama (diccionario de palabras) al N-grama.
				dictNgram = addSentenceToDictionary(words, weight, dictNgram, nSize)
				dictNgramInf = addSentenceToDictionary(wordsInf, weight, dictNgramInf, nSize)

	elif fileFormat == "json": #Caso de arrays JSON
		input_file = open(arg, encoding="utf-8") #Apertura del array JSON con el formato de codificación de caracteres UTF-8, indicado para el uso de caracteres especiales como vocales con tilde o letra Ñ.
		json_array = json.load(input_file) #Carga del array separado por objetos JSON.
		
		for item in json_array: 
			words = replacePunctuation(item[keySentence]).split() #Se obtiene la frase del array JSON, se separa en palabras y se procesan los signos de puntuación (ver explicación en función 'replacePunctuation')
			wordsInf = replaceToInfinitive(words, verbs_array) #Se obtienen las frases pasando todos los verbos a infinitivo
			for w in words: #Se incluyen las nuevas palabras en la lista de palabras 
				if w.lower() not in listWords:
					#print(w)
					listWords.append(w.lower())
			for wI in wordsInf: #Se incluyen las nuevas palabras en la lista de palabras que solo incluye infinitivos
				if wI.lower() not in listWordsInf:
					listWordsInf.append(wI.lower())
			weight = 1 #Cada nueva frase se suma con peso 1 a los diccionarios existentes.

			sizeSentence = len(words) #Tamaño de la frase en número de palabras
			if(sizeSentence > 0): #Si la frase no acaba en un signo de puntuación, se le añade uno. En este caso el punto.
				if words[sizeSentence-1] not in finishOperators:
					words.append('.')
					wordsInf.append('.')

			for nSize in range(maxNgramSize+1): #Se generan los N-gramas contenidos en la frase que está siendo analizada. Desde el 1-grama (diccionario de palabras) al N-grama.
				dictNgram = addSentenceToDictionary(words, weight, dictNgram, nSize)
				dictNgramInf = addSentenceToDictionary(wordsInf, weight, dictNgramInf, nSize)

	else: #Si el formato de archivo introducido no es válido, se le indica al usuario. Además, se cambia el boolean 'valid' a falso.
		print('Formato no válido. Indique "txt" o "json".')
		valid = False

	if valid: #Se generan los diccionarios del parámetro Backoff tanto para el diccionario regular como para el de infinitivos si se ha proporcionado un archivo de entrada válido
		dictBackoff = generateDictBackoffKatz(dictNgram) 
		dictBackoffInf = generateDictBackoffKatz(dictNgramInf)

	return listWords, listWordsInf, dictNgram, dictNgramInf, dictBackoff, dictBackoffInf #Se devuelven las listas de palabras y diccionarios generados
	